# Full-cross-val workflow
## This code should be valid for individual months and temporal averages
source("https://earth.bsc.es/gitlab/es/sunset/-/raw/dev-test_CERISE/GetProbs.R")

full_crossval_anomalies <- function(recipe, data) {
  cross.method <- recipe$Analysis$Skill$cross.method
  # TODO move check
  if (is.null(cross.method)) {
    cross.method <- 'leave-one-out'
  }
  categories <- recipe$Analysis$Workflow$Probabilities$percentiles
  categories <- lapply(categories, function (x) { 
                       sapply(x, function(y) {
                              round(eval(parse(text = y)),2)})})
  cal.method <- recipe$Analysis$Workflow$Calibration$method
  ncores <- recipe$Analysis$ncores
  na.rm <- recipe$Analysis$remove_NAs 
  ## TODO remove if the recipe is checked:
  na.rm <- TRUE
  ## data dimensions 
  sdate_dim <- dim(data$hcst$data)['syear']
  nmemb <- dim(data$hcst$data)['ensemble']
  nftime <- dim(data$hcst$data)['time']
  # spatial dims
  if ('latitude' %in% names(dim(data$hcst$data))) {
    nlats <- dim(data$hcst$data)['latitude']
    nlons <- dim(data$hcst$data)['longitude']
    agg = 'global'
    ev_dim_names <- c('dat', 'var', 'sday', 'sweek', 'time',
                      'latitude', 'longitude', 'ensemble', 'syear')
    tr_dim_names <- c('dat', 'var', 'sday', 'sweek', 'ensemble', 'time',
                      'latitude', 'longitude', 'unneeded', 'syear')
  } else if  ('region' %in% names(dim(data$hcst$data))) {
    agg = 'region'
    nregions <- dim(data$hcst$data)['region']
    ev_dim_names <- c('dat', 'var', 'sday', 'sweek', 'time',
                      'region', 'ensemble', 'syear')
    tr_dim_names <- c('dat', 'var', 'sday', 'sweek', 'ensemble', 'time',
                      'region', 'unneeded', 'syear')
  }

  # TODO fix it to use new version https://earth.bsc.es/gitlab/external/cstools/-/blob/dev-cross-indices/R/CST_Calibration.R#L570
  cross <- CSTools:::.make.eval.train.dexes(eval.method = cross.method,
                                            amt.points = sdate_dim,
                                            amt.points_cor = NULL) # k = ?
  outdim <- length(cross) # leave-one-out should be equal to sdate_dim

  ## output objects
  ano_hcst_ev_res <- NULL
  ano_obs_ev_res <- NULL
  ano_obs_tr_res <- NULL
  # as long as probs requested in recipe:
  lims_ano_hcst_tr_res <- lapply(categories, function(X) {NULL})
  lims_ano_obs_tr_res <- lapply(categories, function(X) {NULL})

  fcst_probs <- lapply(categories, function(x){NULL})
  hcst_probs_ev <- lapply(categories, function(x){NULL})
  obs_probs_ev <- lapply(categories, function(x){NULL})

  for (t in 1:outdim) {
    info(recipe$Run$logger, paste("crossval:", t))

    # subset years: Subset works at BSC not at Athos
    ## training indices
    obs_tr <- Subset(data$obs$data, along = 'syear',
   		     indices = cross[[t]]$train.dexes)
    hcst_tr <- Subset(data$hcst$data, along = 'syear', 
  		      indices = cross[[t]]$train.dexes)
    ## evaluation indices
    hcst_ev <- Subset(data$hcst$data, along = 'syear',
  		      indices = cross[[t]]$eval.dexes, drop = 'selected')
    obs_ev <- Subset(data$obs$data, along = 'syear',
  		     indices = cross[[t]]$eval.dexes, drop = 'selected')
    # compute climatology:
    clim_obs_tr <- MeanDims(obs_tr, 'syear')
    clim_hcst_tr <- MeanDims(hcst_tr, c('syear', 'ensemble'))
    # compute anomalies:
    ano_obs_tr <- s2dv::Ano(obs_tr, clim_obs_tr,
                            ncores = ncores)
    ano_hcst_tr <- s2dv::Ano(hcst_tr, clim_hcst_tr,
                             ncores = ncores)
    ano_hcst_ev <- s2dv::Ano(hcst_ev, clim_hcst_tr,
                             ncores = ncores)
    ano_obs_ev <- s2dv::Ano(obs_ev, clim_obs_tr,
                            ncores = ncores)
    # compute category limits
    lims_ano_hcst_tr <- Apply(ano_hcst_tr, target_dims = c('syear', 'ensemble'),
                              fun = function(x, prob_lims) {
                                lapply(prob_lims, function(ps) {
                                  quantile(as.vector(x), ps, na.rm = na.rm)})},
                              output_dims = lapply(categories, function(x) {'cat'}), 
                              prob_lims = categories,
                              ncores = ncores)
    lims_ano_obs_tr <- Apply(ano_obs_tr, target_dims = c('syear'),#, 'ensemble'),
                             fun = function(x, prob_lims) {
                               lapply(prob_lims, function(ps) {
                                 quantile(as.vector(x), ps, na.rm = na.rm)})},
                             output_dims = lapply(categories, function(x){'cat'}),
                             prob_lims = categories,
                             ncores = ncores)
    #store results
    ano_hcst_ev_res <- abind(ano_hcst_ev_res, ano_hcst_ev,
                             along = length(dim(ano_hcst_ev)) + 1)
    ano_obs_ev_res <- abind(ano_obs_ev_res, ano_obs_ev,
                            along = length(dim(ano_obs_ev)) + 1)
    ano_obs_tr_res <- abind(ano_obs_tr_res, ano_obs_tr,
                            along = length(dim(ano_obs_tr)) + 1)
    for(ps in 1:length(categories)) {
      lims_ano_hcst_tr_res[[ps]] <- abind(lims_ano_hcst_tr_res[[ps]], lims_ano_hcst_tr[[ps]],
                                along = length(dim(lims_ano_hcst_tr[[ps]])) + 1)
      lims_ano_obs_tr_res[[ps]] <- abind(lims_ano_obs_tr_res[[ps]], lims_ano_obs_tr[[ps]],
                                 along = length(dim(lims_ano_obs_tr[[ps]])) + 1)
    }
  }
  info(recipe$Run$logger,
       "#### Anomalies Cross-validation loop ended #####")
  gc()

  # Add dim names:
  names(dim(ano_hcst_ev_res)) <- ev_dim_names
  names(dim(ano_obs_ev_res)) <- ev_dim_names
  names(dim(ano_obs_tr_res)) <- tr_dim_names
  # To make crps_clim to work the reference forecast need to have same dims as obs:
  ano_obs_tr_res <- Subset(ano_obs_tr_res, along = 'unneeded', indices = 1, drop = 'selected')
  for(ps in 1:length(categories)) {
    names(dim(lims_ano_hcst_tr_res[[ps]])) <- c('cat', 'dat', 'var', 'sday', 'sweek',
                                                'time', 'latitude', 'longitude', 'syear')
    names(dim(lims_ano_obs_tr_res[[ps]])) <- c('cat', 'dat', 'var', 'sday', 'sweek',
                                               'time', 'latitude', 'longitude', 'extra', 'syear')
    lims_ano_obs_tr_res[[ps]] <- Subset(lims_ano_obs_tr_res[[ps]],
                                        along = 'extra', indices = 1, drop = 'selected')
  }

  # Forecast anomalies:
  if (!is.null(data$fcst)) {
    clim_hcst <- Apply(ano_hcst_ev_res,
                       target_dims = c('syear', 'ensemble'),
                       mean,
                       na.rm = na.rm,
                       ncores = ncores)$output1
    data$fcst$data <- Ano(data = data$fcst$data, clim = clim_hcst)
    # Terciles limits using the whole hindcast period:
    lims_fcst <- Apply(ano_hcst_ev_res, target_dims = c('syear', 'ensemble'),
                       fun = function(x, prob_lims) {
                         lapply(prob_lims, function(ps) {
                         quantile(as.vector(x), ps, na.rm = na.rm)})},
                       output_dims = lapply(categories, function(x) {'cat'}),
                       prob_lims = categories,
                       ncores = ncores)
  }

  # Compute Probabilities
  for (ps in 1:length(categories)) {
    hcst_probs_ev[[ps]] <- GetProbs(ano_hcst_ev_res, time_dim = 'syear',
                              prob_thresholds = NULL,
                              bin_dim_abs = 'cat',
                              indices_for_quantiles = NULL,
                              memb_dim = 'ensemble', abs_thresholds = lims_ano_hcst_tr[[ps]],
                              ncores = ncores)
    obs_probs_ev[[ps]] <- GetProbs(ano_obs_ev_res, time_dim = 'syear',
                                   prob_thresholds = NULL,
                                   bin_dim_abs = 'cat',
                                   indices_for_quantiles = NULL,
                                   memb_dim = 'ensemble',
                                   abs_thresholds = lims_ano_obs_tr_res[[ps]],
                                   ncores = ncores)
    if (!is.null(data$fcst)) {
      fcst_probs[[ps]] <- GetProbs(data$fcst$data, time_dim = 'syear',
                                   prob_thresholds = NULL,
                                   bin_dim_abs = 'cat',
                                   indices_for_quantiles = NULL,
                                   memb_dim = 'ensemble',
                                   abs_thresholds = lims_fcst[[ps]],
                                   ncores = ncores)
    }
  }
  # Convert to s2dv_cubes the resulting anomalies
  ano_hcst <- data$hcst
  ano_hcst$data <- ano_hcst_ev_res 
  ano_obs <- data$obs
  ano_obs$data <- ano_obs_ev_res

  info(recipe$Run$logger, 
       "#### Anomalies  and Probabilities Done #####") 
  if (recipe$Analysis$Workflow$Anomalies$save != 'none') {
    info(recipe$Run$logger, "##### START SAVING ANOMALIES #####")
    recipe$Run$output_dir <- paste0(recipe$Run$output_dir,
                                    "/outputs/Anomalies/")
    # Save forecast
    if ((recipe$Analysis$Workflow$Anomalies$save %in% 
           c('all', 'exp_only', 'fcst_only')) && !is.null(data$fcst)) {
      save_forecast(recipe = recipe, data_cube = data$fcst, type = 'fcst')
    }
    # Save hindcast
    if (recipe$Analysis$Workflow$Anomalies$save %in%
        c('all', 'exp_only')) {
      save_forecast(recipe = recipe, data_cube = ano_hcst, type = 'hcst')
    }
    # Save observation
    if (recipe$Analysis$Workflow$Anomalies$save == 'all') {
      save_observations(recipe = recipe, data_cube = ano_obs)
    }
  }
  # Save probability bins
  result <- list()
  all_names <- NULL
  for (ps in 1:length(categories)) {
    for (perc in 1:(length(categories[[ps]]) + 1)) {
#ps <- 1
      if (perc == 1) {
        name_elem <- paste0("below_", categories[[ps]][perc])
      } else if (perc == length(categories[[ps]]) + 1) {
        name_elem <- paste0("above_", categories[[ps]][perc-1])
      } else {
        name_elem <- paste0("between_", categories[[ps]][perc-1], 
                            "_and_", categories[[ps]][perc])
      }
      result <- append(list(Subset(hcst_probs_ev[[ps]],
                       along = 'cat', indices = ps, drop = 'all')), result)
      all_names <- c(all_names, name_elem)
    }
  }
  names(result) <- all_names
  if (!('var' %in% names(dim(result[[1]])))) {
    result <- lapply(result, function(x) {
                     dim(x) <- c(var = 1, dim(x))
                     return(x)})
  }
  if (recipe$Analysis$Workflow$Probabilities$save %in%
    c('all', 'bins_only')) {
    save_probabilities(recipe = recipe, probs = result,
                       data_cube = data$hcst, agg = agg,
                       type = "hcst")
    # TODO Forecast
    probs_fcst <- NULL
    if (!is.null(probs_fcst)) {
      save_probabilities(recipe = recipe, probs = probs_fcst,
                         data_cube = data$fcst, agg = agg, 
                         type = "fcst")
    }
  }
  # Save ensemble mean for multimodel option:
  hcst_EM <- MeanDims(ano_hcst$data, 'ensemble', drop = T)  
  save_metrics(recipe = recipe, metrics = list(hcst_EM = hcst_EM),
               data_cube = data$hcst, agg = agg,
               module = "statistics")
  if (!is.null(data$fcst)) {
    fcst_EM <- MeanDims(data$fcst$data, 'ensemble', drop = T)
    save_metrics(recipe = recipe, metrics = list(fcst_EM = fcst_EM),
                 data_cube = data$fcst, agg = agg,
                 module = "statistics")
  }
  return(list(hcst = ano_hcst, obs = ano_obs, fcst = data$fcst,
              hcst.full_val = data$hcst, obs.full_val = data$obs,
              cat_lims = list(hcst_tr = lims_ano_hcst_tr_res,
                              obs_tr = lims_ano_obs_tr_res),
              probs = list(hcst_ev = hcst_probs_ev,
                           obs_ev = obs_probs_ev),
              ref_obs_tr = ano_obs_tr_res))
}

